#!/usr/bin/python3

# Create gcode file for gear rack in low warping mode
# Details on:
# https://homofaciens.de/technics-machines-3D-printer-Longer-LK5Pro-warping_en.htm

import os
import math

fileName = "Gear-Rack-calculated.gcode"

X = 0.0
Y = 0.0
Z = 0.0
E = 0.0

Xmax = 300.0
Ymax = 300.0
Zmax = 400.0

F_travel = 7800.0
F_1st_layer = 600
F_fill = 1800
F_print = 1800
F_retract = 6000
Retract = 3.000
Bed_temp = 100
Nozzle_temp = 240

rack_l = 399.000
rack_w = 20.000
rack_h = 3.000
rack_wall = 2.000

E_width = 0.600
# Extrusion width 0.5mm, data from box-ABS.gcode, created with Slic3r
# In file line 586 to 591:
# Extruder goes from E665.97304 to E669.65044
# while X movement from X198.379 to X101.621
# Extruder movement per mm X movement per mm layer height (0.2mm was used in example file):
ExtrusionRate = (((669.65044 - 665.97304) / (198.379 - 101.621)) / 0.2) / 0.5
ExtrusionRate = ExtrusionRate * 1.05 # Filament factor


LAYER_H_1st = 0.350
LAYER_H = 0.200

# LAYERS must be an even number!
# Height of base plate is LAYER_H_1st + (LAYERS - 1) * LAYER_H
LAYERS = 10


########################################################################
def driveToXYZ(driveX, driveY, driveZ, retract):
    global E

    # retract 0 = no retract
    # retract 1 = only retract
    # retract 2 = only forward
    # retract 3 = retract at beginning, forward at end of movement

    if retract == 1 or retract == 3:
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      storeFile.write("G1 E" + format(-Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      E = 0
    storeFile.write("G1 F" + format(F_travel, ".3f") + "\n")
    storeFile.write("G1 Z" + format(driveZ, ".3f") + "\n")
    storeFile.write("G1 X" + format(driveX, ".3f") + " Y" + format(driveY, ".3f") + "\n")
    if retract == 2 or retract == 3:
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      storeFile.write("G1 E" + format(Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
      storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
      E = 0


########################################################################
def extrudeToXYZ(eXfrom, eYfrom, eZfrom, eXto, eYto, eZto, layer_h, E_width, E, F = 0.0, makeDots = False):
    global ExtrusionRate
    global dotDistance
    global currentDot

    travel_X = math.pow(eXfrom - eXto, 2)
    travel_Y = math.pow(eYfrom - eYto, 2)
    travel_Z = math.pow(eZfrom - eZto, 2)
    travel_l = math.sqrt(travel_X + travel_Y + travel_Z)
    # Eold = E

    if F != 0.0:
      storeFile.write("G1 F" + format(F, ".3f") + "\n")

    X1 = eXfrom
    Y1 = eYfrom
    Z1 = eZfrom

    if makeDots:
      ratioX = (eXto - eXfrom) / travel_l
      ratioY = (eYto - eYfrom) / travel_l
      ratioZ = (eZto - eZfrom) / travel_l
      while travel_l > currentDot:
        X2 = X1 + ratioX * currentDot
        Y2 = Y1 + ratioY * currentDot
        Z2 = Z1 + ratioZ * currentDot
        E = E + currentDot * ExtrusionRate * layer_h * E_width
        storeFile.write("G1 X" + format(X2, ".3f") + " Y" + format(Y2, ".3f") + " Z" + format(Z2, ".3f") + " E" + format(E, ".5f") + "\n")
        Ztop = Z2
        Z2 = Z2 - Z_HUB
        if Z2 < 0.1:
          Z2 = 0.1
        storeFile.write("G1 Z" + format(Z2, ".3f") + "\n")
        Z2 = Ztop
        storeFile.write("G1 Z" + format(Z2, ".3f") + "\n")
        travel_l = travel_l - currentDot
        currentDot = dotDistance
        X1 = X2
        Y1 = Y2
        Z1 = Z1
      if travel_l > 0.0:
        X2 = X1 + ratioX * travel_l
        Y2 = Y1 + ratioY * travel_l
        Z2 = Z1 + ratioZ * travel_l
        E = E + travel_l * ExtrusionRate * layer_h * E_width
        storeFile.write("G1 X" + format(X2, ".3f") + " Y" + format(Y2, ".3f") + " Z" + format(Z2, ".3f") + " E" + format(E, ".5f") + "\n")
        currentDot = currentDot - travel_l


    else:
      E = E + travel_l * ExtrusionRate * layer_h * E_width
      # print("travel_l = ", travel_l, ", E = ", E, ", Eold = ", Eold)
      storeFile.write("G1 X" + format(eXto, ".3f") + " Y" + format(eYto, ".3f") + " Z" + format(eZto, ".3f") + " E" + format(E, ".5f") + "\n")

    return E



########################################################################
########################################################################
########################################################################


storeFile = open(fileName, "w")

storeFile.write("M190 S" + format(Bed_temp, ".3f") + " ; set bed temperature and wait for it to be reached" + "\n")
storeFile.write("M104 S" + format(Nozzle_temp, ".3f") + " ; set temperature" + "\n")
storeFile.write("G0 Z 10.000 ; Lift Z axis" + "\n")
storeFile.write("G28 ; home all axes" + "\n")
storeFile.write("G1 Z5 F5000 ; lift nozzle" + "\n")
storeFile.write("M109 S" + format(Nozzle_temp, ".3f") + " ; set temperature and wait for it to be reached" + "\n")
storeFile.write("G21 ; set units to millimeters" + "\n")
storeFile.write("G90 ; use absolute coordinates" + "\n")
storeFile.write("M82 ; use absolute distances for extrusion" + "\n")
storeFile.write("G92 E0 ; Set current extruder position to Zero" + "\n")
storeFile.write("G1 Z10.000 F" + format(F_travel, ".3f") + "\n")
storeFile.write("G1 E10.000 F120.000" + "\n")
storeFile.write("G92 E0" + "\n")
storeFile.write("G1 E" + format(-Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
storeFile.write("G1 Z" + format(LAYER_H_1st, ".3f") + " F" + format(F_travel, ".3f") + "\n")
storeFile.write("G92 E0" + "\n")



E = 0
PI = 3.1415927
alpha = 45*PI/180-math.atan((rack_w/2)/(rack_l/2))
print("alpha = ", alpha * 180 / PI)
rack_dia = math.sqrt((rack_w/2) * (rack_w/2) + (rack_l/2) * (rack_l/2))
Xstart = Xmax / 2 - math.cos(alpha) * rack_dia
Ystart = Ymax / 2 + math.sin(alpha) * rack_dia


storeFile.write("G1 X" + format(Xstart, ".3f") + " Y" + format(Ystart, ".3f") + " F" + format(F_travel, ".3f") + "\n")
storeFile.write("G1 E" + format(Retract, ".5f") + " F" + format(F_retract, ".3f") + "\n")
storeFile.write("G92 E0" + "\n")
E = 0
storeFile.write("G1 F" + format(F_travel, ".3f") + "\n")

py_edge = 15.0
delta_edge = py_edge + rack_wall - E_width
# First square to prepare nozzle
X = Xstart + rack_wall
Y = Ystart + rack_wall - 50
Z = LAYER_H_1st


# Extrude around maximum points
eXf = Xstart - 3.50
eYf = Ystart
eXto = eXf + (rack_l + 5.0) / math.sqrt(2)
eYto = eYf - (rack_l + 5.0) / math.sqrt(2)
driveToXYZ(eXf, eYf, Z, 3)
E = 0
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)

eXf = eXto
eYf = eYto
eXto = eXf + (rack_w + 5.0) / math.sqrt(2)
eYto = eYf + (rack_w + 5.0) / math.sqrt(2)
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)

eXf = eXto
eYf = eYto
eXto = eXf - (rack_l + 5.0) / math.sqrt(2)
eYto = eYf + (rack_l + 5.0) / math.sqrt(2)
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)

eXf = eXto
eYf = eYto
eXto = Xstart - 1.0
eYto = Ystart + (rack_w / 2) * math.sqrt(2) + 1.0
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)

eXf = eXto
eYf = eYto
eXto = Xstart - 3.50
eYto = Ystart
E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, LAYER_H_1st, E_width, E, F_1st_layer, False)


X = Xstart
Y = Ystart
Z = LAYER_H_1st


# Create socket
layer_h = LAYER_H
F = F_1st_layer
Z = LAYER_H_1st
doRetract = 3
firstFill = True

a = math.sqrt((rack_w/2) * (rack_w/2) * 2)
e2 = math.sqrt(E_width * E_width / 2)
#e2 = E_width # !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
print("a =", a)
for Layer in range(LAYERS):
  Z = LAYER_H_1st + Layer * LAYER_H
  F = F_1st_layer
  layer_h = LAYER_H_1st
  if Z > LAYER_H_1st:
    layer_h = LAYER_H
    F = F_print

  retract = 3
  for j in range(Layer, int((rack_w / math.sqrt(2 * E_width * E_width))/2)):
    e = a - j * 2 * E_width
    eXf = Xstart + j * E_width
    eYf = Ystart + j * E_width
    eXto = eXf + e
    eYto = eYf
    driveToXYZ(eXf, eYf, Z, retract)
    if retract == 3:
      retract = 0
      E = 0
    E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)
    eXf = eXto
    eYf = eYto
    eXto = eXf
    eYto = eYf + e
    E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)

    if Layer == 0: # Initial square for bed adhesion
      eXf = eXto
      eYf = eYto
      eXto = eXf - e
      eYto = eYf
      driveToXYZ(eXf, eYf, Z, retract)
      if retract == 3:
        retract = 0
        E = 0
      E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)
      eXf = eXto
      eYf = eYto
      eXto = eXf
      eYto = eYf - e
      E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)


growEnd = int((rack_l - rack_w/2) / (math.sqrt(2) * E_width)) + 1
for grow in range(growEnd):
  retract = 3
  for Layer in range(0, LAYERS, 2):
    Z = LAYER_H_1st + Layer * LAYER_H
    F = F_1st_layer
    layer_h = LAYER_H_1st
    if Z > LAYER_H_1st:
      layer_h = LAYER_H
      F = F_print


    e = a
    eXf = Xstart + grow * E_width - Layer * E_width
    eYf = Ystart - grow * E_width + Layer * E_width
    eXto = eXf + e
    eYto = eYf
    Lcut = Layer * E_width * 2 - grow * E_width * 2
    if Lcut < 0:
      Lcut = 0

    if retract == 3:
      driveToXYZ(eXf + Lcut, eYf, Z + LAYERS * LAYER_H, 1)
      driveToXYZ(eXf + Lcut, eYf, Z, 2)
      retract = 0
      E = 0
    else:
      driveToXYZ(eXf + Lcut, eYf, Z, 0)
    E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)
    eXf = eXto
    eYf = eYto
    eXto = eXf
    eYto = eYf + e
    E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto - Lcut, Z, layer_h, E_width, E, F, False)


    Z = LAYER_H_1st + (Layer + 1) * LAYER_H
    F = F_print
    Lcut = (Layer + 1) * E_width * 2 - grow * E_width * 2
    if Lcut < 0:
      Lcut = 0

    eXf = eXto - E_width
    eYf = eYto + E_width
    eXto = eXf
    eYto = eYf - e
    driveToXYZ(eXf, eYf - Lcut, Z, retract)
    E = extrudeToXYZ(eXf, eYf - Lcut, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)
    eXf = eXto
    eYf = eYto
    eXto = eXf - e
    eYto = eYf
    E = extrudeToXYZ(eXf, eYf, Z, eXto + Lcut, eYto, Z, layer_h, E_width, E, F, False)


retract = 3
for grow in range(growEnd, growEnd + int((rack_w/2) / (E_width * math.sqrt(2))) + 2, 2):

  Z = LAYER_H_1st
  F = F_1st_layer
  layer_h = LAYER_H_1st

  e = a
  eXf = Xstart + grow * E_width
  eYf = Ystart - grow * E_width
  eXto = eXf + e
  eYto = eYf

  if retract == 3:
    driveToXYZ(eXf + Lcut, eYf, Z + LAYERS * LAYER_H, 1)
    driveToXYZ(eXf + Lcut, eYf, Z, 2)
    retract = 0
    E = 0
  else:
    driveToXYZ(eXf + Lcut, eYf, Z, 0)
  E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)
  eXf = eXto
  eYf = eYto
  eXto = eXf
  eYto = eYf + e
  E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)

  eXf = eXto + E_width
  eYf = eYto - E_width
  eXto = eXf
  eYto = eYf - e
  driveToXYZ(eXf, eYf, Z, retract)
  E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)
  eXf = eXto
  eYf = eYto
  eXto = eXf - e
  eYto = eYf
  E = extrudeToXYZ(eXf, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)



# End of bar
retract = 3
for Layer in range(1, LAYERS):# Left
  Z = LAYER_H_1st + Layer * LAYER_H
  F = F_1st_layer
  layer_h = LAYER_H_1st
  if Z > LAYER_H_1st:
    layer_h = LAYER_H
    F = F_print

  for j in range(0, int((rack_w/2) / E_width), 2):
    eXf = Xstart + growEnd * E_width + j * e2 - Layer * E_width
    eYf = Ystart - growEnd * E_width - j * e2 + Layer * E_width
    eXto = eXf + e2 * j
    eYto = eYf + e2 * j
    if retract == 3:
      driveToXYZ(eXf + Lcut, eYf, Z + LAYERS * LAYER_H, 1)
      driveToXYZ(eXf + Lcut, eYf, Z, 2)
      retract = 0
      E = 0
    else:
      driveToXYZ(eXf + Lcut, eYf, Z, 0)
    E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)
    eXto = Xstart + growEnd * E_width + (j + 1) * e2 - Layer * E_width
    eYto = Ystart - growEnd * E_width - (j + 1) * e2 + Layer * E_width
    eXf = eXto + e2 * (j + 1)
    eYf = eYto + e2 * (j + 1)

    driveToXYZ(eXf + Lcut, eYf, Z, 0)
    E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)

retract = 3
for Layer in range(1, LAYERS): # Right
  Z = LAYER_H_1st + Layer * LAYER_H
  F = F_1st_layer
  layer_h = LAYER_H_1st
  if Z > LAYER_H_1st:
    layer_h = LAYER_H
    F = F_print

  for j in range(0, int((rack_w/2) / E_width), 2):
    eXf = Xstart + growEnd * E_width + j * e2 - Layer * E_width + a
    eYf = Ystart - growEnd * E_width - j * e2 + Layer * E_width + a
    eXto = eXf - e2 * j
    eYto = eYf - e2 * j
    if retract == 3:
      driveToXYZ(eXf + Lcut, eYf, Z + LAYERS * LAYER_H, 1)
      driveToXYZ(eXf + Lcut, eYf, Z, 2)
      retract = 0
      E = 0
    else:
      driveToXYZ(eXf + Lcut, eYf, Z, 0)
    E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)

    eXto = Xstart + growEnd * E_width + (j + 1) * e2 - Layer * E_width + a
    eYto = Ystart - growEnd * E_width - (j + 1) * e2 + Layer * E_width + a
    eXf = eXto - e2 * (j + 1)
    eYf = eYto - e2 * (j + 1)

    driveToXYZ(eXf + Lcut, eYf, Z, 0)
    E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)

retract = 3
for Layer in range(1, LAYERS): # End of end
  Z = LAYER_H_1st + Layer * LAYER_H
  F = F_1st_layer
  layer_h = LAYER_H_1st
  if Z > LAYER_H_1st:
    layer_h = LAYER_H
    F = F_print
  for j in range(0, int(Layer * math.sqrt(2)) + 2, 2):
    eXf = Xstart + growEnd * E_width + math.sqrt((rack_w/2)*(rack_w/2)/2) - (j + 0.5) * e2
    eYf = Ystart - growEnd * E_width - math.sqrt((rack_w/2)*(rack_w/2)/2) + (j + 0.5) * e2
    eXto = eXf + math.sqrt(rack_w * rack_w / 2)
    eYto = eYf + math.sqrt(rack_w * rack_w / 2)
    if retract == 3:
      driveToXYZ(eXf + Lcut, eYf, Z + LAYERS * LAYER_H, 1)
      driveToXYZ(eXf + Lcut, eYf, Z, 2)
      retract = 0
      E = 0
    else:
      driveToXYZ(eXf + Lcut, eYf, Z, 0)
    E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)

    if j + 1 < int(Layer * math.sqrt(2)) + 2:
      eXto = Xstart + growEnd * E_width + math.sqrt((rack_w/2)*(rack_w/2)/2) - (j + 1.5) * e2
      eYto = Ystart - growEnd * E_width - math.sqrt((rack_w/2)*(rack_w/2)/2) + (j + 1.5) * e2
      eXf = eXto + math.sqrt(rack_w * rack_w / 2)
      eYf = eYto + math.sqrt(rack_w * rack_w / 2)
      driveToXYZ(eXf + Lcut, eYf, Z, 0)
      E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_width, E, F, False)


# Gear teeth
g_w = 3.0
g_l = rack_w / 2
g_h = 3
g_no = int(rack_l / g_w)
g_h_layers = int(g_h / layer_h)
E_width_new = 0.5
for Layer in range(g_h_layers-7):
  Z = LAYER_H_1st + (LAYERS + Layer) * LAYER_H
  gap = g_w / g_h_layers * (Layer + 3)
  gap_e = gap / math.sqrt(2)
  for g in range(g_no):
    g_w_layers = int((g_w - gap) / E_width_new)
    E_new = (g_w - gap) / g_w_layers
    retract = 3
    for j in range(0, g_w_layers, 2):
      eXf = Xstart + g * math.sqrt(g_w*g_w/2) + (gap_e / 2) + j * E_new / math.sqrt(2)
      eYf = Ystart - g * math.sqrt(g_w*g_w/2) - (gap_e / 2) - j * E_new / math.sqrt(2)
      eXto = eXf + math.sqrt(g_l*g_l/2)
      eYto = eYf + math.sqrt(g_l*g_l/2)
      if retract == 3:
        driveToXYZ(eXf + Lcut, eYf, Z, 3)
        retract = 0
        E = 0
      else:
        driveToXYZ(eXf + Lcut, eYf, Z, 0)
      E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_new, E, F, False)

      if j + 1 < g_w_layers:
        eXto = Xstart + g * math.sqrt(g_w*g_w/2) + (gap_e / 2) + (j + 1) * E_new / math.sqrt(2)
        eYto = Ystart - g * math.sqrt(g_w*g_w/2) - (gap_e / 2) - (j + 1) * E_new / math.sqrt(2)
        eXf = eXto + math.sqrt(g_l*g_l/2)
        eYf = eYto + math.sqrt(g_l*g_l/2)
        driveToXYZ(eXf + Lcut, eYf, Z, 0)
        E = extrudeToXYZ(eXf + Lcut, eYf, Z, eXto, eYto, Z, layer_h, E_new, E, F, False)




# Final Gcode lines, turn off nozzle and bed et cetera...
storeFile.write("G92 E0" + "\n")
storeFile.write("M104 S0 ; turn off hotend temperature" + "\n")
storeFile.write("M140 S0 ; turn off bed temperature" + "\n")
storeFile.write("G28 X0  ; home X axis" + "\n")
storeFile.write("M84     ; disable motors" + "\n")

print("Gcode file written!")
storeFile.close()


print("...done!")
